@@ -34,7 +34,11 @@ Galois/Counter Mode (GCM).
 
 __all__ = ['GcmMode']
 
-from Crypto.Util.py3compat import byte_string, bord, unhexlify, _copy_bytes
+from binascii import unhexlify
+
+from Crypto.Util.py3compat import bord, _copy_bytes
+
+from Crypto.Util._raw_api import is_buffer
 
 from Crypto.Util.number import long_to_bytes, bytes_to_long
 from Crypto.Hash import BLAKE2s
@@ -44,7 +48,7 @@ from Crypto.Util._raw_api import (load_pycryptodome_raw_lib, VoidPointer,
                                   create_string_buffer, get_raw_buffer,
                                   SmartPointer, c_size_t, c_uint8_ptr)
 
-from Crypto.Util import _cpuid
+from Crypto.Util import _cpu_features
 
 
 # C API by module implementing GHASH
@@ -68,7 +72,7 @@ def _build_impl(lib, postfix):
         imp_funcs = [ getattr(lib, x + "_" + postfix) for x in funcs ]
     except AttributeError:      # Make sphinx stop complaining with its mocklib
         imp_funcs = [ None ] * 3
-    params = dict(list(zip(funcs, imp_funcs)))
+    params = dict(zip(funcs, imp_funcs))
     return GHASH_Imp(**params)
 
 
@@ -83,7 +87,7 @@ _ghash_portable = _get_ghash_portable()
 def _get_ghash_clmul():
     """Return None if CLMUL implementation is not available"""
 
-    if not _cpuid.have_clmul():
+    if not _cpu_features.have_clmul():
         return None
     try:
         api = _ghash_api_template.replace("%imp%", "clmul")
@@ -182,8 +186,9 @@ class GcmMode(object):
 
         if len(nonce) == 0:
             raise ValueError("Nonce cannot be empty")
-        if isinstance(nonce, str):
-            raise TypeError("Nonce must be a byte string")
+        
+        if not is_buffer(nonce):
+            raise TypeError("Nonce must be bytes, bytearray or memoryview")
 
         # See NIST SP 800 38D, 5.2.1.1
         if len(nonce) > 2**64 - 1:
@@ -328,7 +333,7 @@ class GcmMode(object):
         if len_cache > 0:
             self._update(b'\x00' * (16 - len_cache))
 
-    def encrypt(self, plaintext):
+    def encrypt(self, plaintext, output=None):
         """Encrypt data with the key and the parameters set at initialization.
 
         A cipher object is stateful: once you have encrypted a message
@@ -352,9 +357,13 @@ class GcmMode(object):
           plaintext : bytes/bytearray/memoryview
             The piece of data to encrypt.
             It can be of any length.
+        :Keywords:
+          output : bytearray/memoryview
+            The location where the ciphertext must be written to.
+            If ``None``, the ciphertext is returned.
         :Return:
-            the encrypted data, as a byte string.
-            It is as long as *plaintext*.
+          If ``output`` is ``None``, the ciphertext as ``bytes``.
+          Otherwise, ``None``.
         """
 
         if self.encrypt not in self._next:
@@ -362,13 +371,13 @@ class GcmMode(object):
                             " initialization or an update()")
         self._next = [self.encrypt, self.digest]
 
-        ciphertext = self._cipher.encrypt(plaintext)
+        ciphertext = self._cipher.encrypt(plaintext, output=output)
 
         if self._status == MacStatus.PROCESSING_AUTH_DATA:
             self._pad_cache_and_update()
             self._status = MacStatus.PROCESSING_CIPHERTEXT
 
-        self._update(ciphertext)
+        self._update(ciphertext if output is None else output)
         self._msg_len += len(plaintext)
 
         # See NIST SP 800 38D, 5.2.1.1
@@ -377,7 +386,7 @@ class GcmMode(object):
 
         return ciphertext
 
-    def decrypt(self, ciphertext):
+    def decrypt(self, ciphertext, output=None):
         """Decrypt data with the key and the parameters set at initialization.
 
         A cipher object is stateful: once you have decrypted a message
@@ -401,8 +410,13 @@ class GcmMode(object):
           ciphertext : bytes/bytearray/memoryview
             The piece of data to decrypt.
             It can be of any length.
-
-        :Return: the decrypted data (byte string).
+        :Keywords:
+          output : bytearray/memoryview
+            The location where the plaintext must be written to.
+            If ``None``, the plaintext is returned.
+        :Return:
+          If ``output`` is ``None``, the plaintext as ``bytes``.
+          Otherwise, ``None``.
         """
 
         if self.decrypt not in self._next:
@@ -417,7 +431,7 @@ class GcmMode(object):
         self._update(ciphertext)
         self._msg_len += len(ciphertext)
 
-        return self._cipher.decrypt(ciphertext)
+        return self._cipher.decrypt(ciphertext, output=output)
 
     def digest(self):
         """Compute the *binary* MAC tag in an AEAD mode.
@@ -510,22 +524,29 @@ class GcmMode(object):
 
         self.verify(unhexlify(hex_mac_tag))
 
-    def encrypt_and_digest(self, plaintext):
+    def encrypt_and_digest(self, plaintext, output=None):
         """Perform encrypt() and digest() in one step.
 
         :Parameters:
           plaintext : bytes/bytearray/memoryview
             The piece of data to encrypt.
+        :Keywords:
+          output : bytearray/memoryview
+            The location where the ciphertext must be written to.
+            If ``None``, the ciphertext is returned.
         :Return:
-            a tuple with two byte strings:
+            a tuple with two items:
 
-            - the encrypted data
-            - the MAC
+            - the ciphertext, as ``bytes``
+            - the MAC tag, as ``bytes``
+
+            The first item becomes ``None`` when the ``output`` parameter
+            specified a location for the result.
         """
 
-        return self.encrypt(plaintext), self.digest()
+        return self.encrypt(plaintext, output=output), self.digest()
 
-    def decrypt_and_verify(self, ciphertext, received_mac_tag):
+    def decrypt_and_verify(self, ciphertext, received_mac_tag, output=None):
         """Perform decrypt() and verify() in one step.
 
         :Parameters:
@@ -533,14 +554,18 @@ class GcmMode(object):
             The piece of data to decrypt.
           received_mac_tag : byte string
             This is the *binary* MAC, as received from the sender.
-
-        :Return: the decrypted data (byte string).
+        :Keywords:
+          output : bytearray/memoryview
+            The location where the plaintext must be written to.
+            If ``None``, the plaintext is returned.
+        :Return: the plaintext as ``bytes`` or ``None`` when the ``output``
+            parameter specified a location for the result.
         :Raises ValueError:
             if the MAC does not match. The message has been tampered with
             or the key is incorrect.
         """
 
-        plaintext = self.decrypt(ciphertext)
+        plaintext = self.decrypt(ciphertext, output=output)
         self.verify(received_mac_tag)
         return plaintext
 
